"use strict";

require("core-js/modules/es.regexp.exec.js");

require("core-js/modules/es.string.replace.js");

require("core-js/modules/es.object.to-string.js");

require("core-js/modules/es.promise.js");

require("core-js/modules/es.array.concat.js");

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

/*
Copyright 2019-2020 Netfoundry, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 * ZitiWebSocket:
 * 
 * websocket with promise api
 */
var Channel = require('chnl');

var PromiseController = require('promise-controller');

var flatOptions = require('flat-options');

var Requests = require('./requests');

var defaultOptions = require('./options');

var _require = require('../utils/throwif'),
    throwIf = _require.throwIf; // see: https://developer.mozilla.org/en-US/docs/Web/API/WebSocket#Ready_state_constants


var STATE = {
  CONNECTING: 0,
  OPEN: 1,
  CLOSING: 2,
  CLOSED: 3
};
/**
 * @typicalname zws
 */

var ZitiWebSocket = /*#__PURE__*/function () {
  /**
   * Constructor. It does not immediately open connection.
   * Call `open()` method to connect.
   *
   * @param {String} url WebSocket URL
   * @param {Options} [options]
   */
  function ZitiWebSocket(url, options) {
    _classCallCheck(this, ZitiWebSocket);

    this._url = url.replace("ws:", "wss:");
    this._options = flatOptions(options, defaultOptions);
    this._requests = new Requests();
    this._ws = null;
    this._wsSubscription = null;

    this._createOpeningController();

    this._createClosingController();

    this._createChannels();

    this._ctx = options.ctx;
  }
  /**
   * Returns original WebSocket instance created by `options.createWebSocket`.
   *
   * @returns {WebSocket}
   */


  _createClass(ZitiWebSocket, [{
    key: "ws",
    get: function get() {
      return this._ws;
    }
    /**
     * Is WebSocket connection in opening state.
     *
     * @returns {Boolean}
     */

  }, {
    key: "isOpening",
    get: function get() {
      return Boolean(this._ws && this._ws.readyState === STATE.CONNECTING);
    }
    /**
     * Is WebSocket connection opened.
     *
     * @returns {Boolean}
     */

  }, {
    key: "isOpened",
    get: function get() {
      return Boolean(this._ws && this._ws.readyState === STATE.OPEN);
    }
    /**
     * Is WebSocket connection in closing state.
     *
     * @returns {Boolean}
     */

  }, {
    key: "isClosing",
    get: function get() {
      return Boolean(this._ws && this._ws.readyState === STATE.CLOSING);
    }
    /**
     * Is WebSocket connection closed.
     *
     * @returns {Boolean}
     */

  }, {
    key: "isClosed",
    get: function get() {
      return Boolean(!this._ws || this._ws.readyState === STATE.CLOSED);
    }
    /**
     * Event channel triggered when connection to Edge Router is opened.
     *
     * @example
     * zws.onOpen.addListener(() => console.log('Connection opened'));
     *
     * @returns {Channel}
     */

  }, {
    key: "onOpen",
    get: function get() {
      return this._onOpen;
    }
    /**
     * Event channel triggered every time when message is sent to Edge Router.
     *
     * @example
     * zws.onSend.addListener(data => console.log('Message sent', data));
     *
     * @returns {Channel}
     */

  }, {
    key: "onSend",
    get: function get() {
      return this._onSend;
    }
    /**
     * Event channel triggered every time when message received from Edge Router.
     *
     * @example
     * zws.onMessage.addListener(message => console.log(message));
     *
     * @returns {Channel}
     */

  }, {
    key: "onMessage",
    get: function get() {
      return this._onMessage;
    }
    /**
     * Event channel triggered every time received message is successfully unpacked.
     * For example, if you are using JSON transport, the listener will receive already JSON parsed data.
     *
     * @example
     * zws.onUnpackedMessage.addListener(data => console.log(data.foo));
     *
     * @returns {Channel}
     */

  }, {
    key: "onUnpackedMessage",
    get: function get() {
      return this._onUnpackedMessage;
    }
    /**
     * Event channel triggered every time response to some request comes in.
     * Received message considered a response if requestId is found in it.
     *
     * @example
     * zws.onResponse.addListener(data => console.log(data));
     *
     * @returns {Channel}
     */

  }, {
    key: "onResponse",
    get: function get() {
      return this._onResponse;
    }
    /**
     * Event channel triggered when connection closed.
     * Listener accepts single argument `{code, reason}`.
     *
     * @example
     * zws.onClose.addListener(event => console.log(`Connections closed: ${event.reason}`));
     *
     * @returns {Channel}
     */

  }, {
    key: "onClose",
    get: function get() {
      return this._onClose;
    }
    /**
     * Event channel triggered when by ZitiWebsocket 'error' event.
     *
     * @example
     * zws.onError.addListener(event => console.error(event));
     *
     * @returns {Channel}
     */

  }, {
    key: "onError",
    get: function get() {
      return this._onError;
    }
    /**
     * Opens ZitiWebSocket connection. If connection already opened, promise will be resolved with "open event".
     *
     * @returns {Promise<Event>}
     */

  }, {
    key: "open",
    value: function open() {
      var _this = this;

      if (this.isClosing) {
        return Promise.reject(new Error("Can't open WebSocket while closing."));
      }

      if (this.isOpened) {
        return this._opening.promise;
      }

      return this._opening.call(function () {
        _this._opening.promise.catch(function (e) {
          return _this._cleanup(e);
        });

        _this._createWS();
      });
    }
    /**
     * Performs request and waits for response.
     *
     * @param {*} data
     * @param {Object} [options]
     * @param {String|Number} [options.requestId=<auto-generated>]
     * @param {Number} [options.timeout=0]
     * @returns {Promise}
     */

  }, {
    key: "sendRequest",
    value: function sendRequest(data) {
      var _this2 = this;

      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      var requestId = options.requestId || "".concat(Math.random());
      var timeout = options.timeout !== undefined ? options.timeout : this._options.timeout;
      return this._requests.create(requestId, function () {
        _this2._assertRequestIdHandlers();

        var finalData = _this2._options.attachRequestId(data, requestId);

        _this2.sendPacked(finalData);
      }, timeout);
    }
    /**
     * Packs data with `options.packMessage` and sends to the server.
     *
     * @param {*} data
     */

  }, {
    key: "sendPacked",
    value: function sendPacked(data) {
      this._assertPackingHandlers();

      var message = this._options.packMessage(data);

      this.send(message);
    }
    /**
     * Sends data without packing.
     *
     * @param {String|Blob|ArrayBuffer} data
     */

  }, {
    key: "send",
    value: function send(data) {
      throwIf(!this.isOpened, "Can't send data because WebSocket is not opened.");

      this._ctx.logger.debug('zws: send -> data len[%o]', data.byteLength);

      this._ws.send(data);

      this._onSend.dispatchAsync(data);
    }
    /**
     * Closes WebSocket connection. If connection already closed, promise will be resolved with "close event".
     *
     * @returns {Promise<Event>}
     */

  }, {
    key: "close",
    value: function close() {
      var _this3 = this;

      return this.isClosed ? Promise.resolve(this._closing.value) : this._closing.call(function () {
        return _this3._ws.close();
      });
    }
    /**
     * Removes all listeners from zws instance. Useful for cleanup.
     */

  }, {
    key: "removeAllListeners",
    value: function removeAllListeners() {
      this._onOpen.removeAllListeners();

      this._onMessage.removeAllListeners();

      this._onUnpackedMessage.removeAllListeners();

      this._onResponse.removeAllListeners();

      this._onSend.removeAllListeners();

      this._onClose.removeAllListeners();

      this._onError.removeAllListeners();
    }
  }, {
    key: "_createOpeningController",
    value: function _createOpeningController() {
      var connectionTimeout = this._options.connectionTimeout || this._options.timeout;
      this._opening = new PromiseController({
        timeout: connectionTimeout,
        timeoutReason: "Can't open WebSocket within allowed timeout: ".concat(connectionTimeout, " ms.")
      });
    }
  }, {
    key: "_createClosingController",
    value: function _createClosingController() {
      var closingTimeout = this._options.timeout;
      this._closing = new PromiseController({
        timeout: closingTimeout,
        timeoutReason: "Can't close WebSocket within allowed timeout: ".concat(closingTimeout, " ms.")
      });
    }
  }, {
    key: "_createChannels",
    value: function _createChannels() {
      this._onOpen = new Channel();
      this._onMessage = new Channel();
      this._onUnpackedMessage = new Channel();
      this._onResponse = new Channel();
      this._onSend = new Channel();
      this._onClose = new Channel();
      this._onError = new Channel();
    }
  }, {
    key: "_createWS",
    value: function _createWS() {
      var _this4 = this;

      this._ws = this._options.createWebSocket(this._url);
      this._wsSubscription = new Channel.Subscription([{
        channel: this._ws,
        event: 'open',
        listener: function listener(e) {
          return _this4._handleOpen(e);
        }
      }, {
        channel: this._ws,
        event: 'message',
        listener: function listener(e) {
          return _this4._handleMessage(e);
        }
      }, {
        channel: this._ws,
        event: 'error',
        listener: function listener(e) {
          return _this4._handleError(e);
        }
      }, {
        channel: this._ws,
        event: 'close',
        listener: function listener(e) {
          return _this4._handleClose(e);
        }
      }]).on();
    }
  }, {
    key: "_handleOpen",
    value: function _handleOpen(event) {
      this._ctx.logger.debug('zws: _handleOpen: event[%o]', event);

      this._onOpen.dispatchAsync(event);

      this._opening.resolve(event);
    }
  }, {
    key: "_handleMessage",
    value: function _handleMessage(event) {
      this._ctx.logger.debug('zws: _handleMessage: event[%o]', event);

      var data = this._options.extractMessageData(event);

      this._ctx.logger.debug('zws: _handleMessage: recv <- data[%o]', data);

      this._onMessage.dispatchAsync(data);

      this._tryUnpack(data);
    }
  }, {
    key: "_tryUnpack",
    value: function _tryUnpack(data) {
      if (this._options.unpackMessage) {
        data = this._options.unpackMessage(data);

        if (data !== undefined) {
          this._onUnpackedMessage.dispatchAsync(data);

          this._tryHandleResponse(data);
        }
      }
    }
  }, {
    key: "_tryHandleResponse",
    value: function _tryHandleResponse(data) {
      this._ctx.logger.debug('zws: _tryHandleResponse: recv <- data[%o]', data);

      if (this._options.extractRequestId) {
        var requestId = this._options.extractRequestId(data);

        if (requestId) {
          this._onResponse.dispatchAsync(data, requestId);

          this._requests.resolve(requestId, data);
        }
      }
    }
  }, {
    key: "_handleError",
    value: function _handleError(event) {
      this._ctx.logger.error('zws: _handleError: event[%o]', event);

      this._onError.dispatchAsync(event);
    }
  }, {
    key: "_handleClose",
    value: function _handleClose(event) {
      this._ctx.logger.debug('zws: _handleClose: event[%o]', event);

      this._onClose.dispatchAsync(event);

      this._closing.resolve(event);

      var error = new Error("WebSocket closed with reason: ".concat(event.reason, " (").concat(event.code, ")."));

      if (this._opening.isPending) {
        this._opening.reject(error);
      }

      this._cleanup(error);
    }
  }, {
    key: "_cleanupWS",
    value: function _cleanupWS() {
      if (this._wsSubscription) {
        this._wsSubscription.off();

        this._wsSubscription = null;
      }

      this._ws = null;
    }
  }, {
    key: "_cleanup",
    value: function _cleanup(error) {
      this._cleanupWS();

      this._requests.rejectAll(error);
    }
  }, {
    key: "_assertPackingHandlers",
    value: function _assertPackingHandlers() {
      var _this$_options = this._options,
          packMessage = _this$_options.packMessage,
          unpackMessage = _this$_options.unpackMessage;
      throwIf(!packMessage || !unpackMessage, "Please define 'options.packMessage / options.unpackMessage' for sending packed messages.");
    }
  }, {
    key: "_assertRequestIdHandlers",
    value: function _assertRequestIdHandlers() {
      var _this$_options2 = this._options,
          attachRequestId = _this$_options2.attachRequestId,
          extractRequestId = _this$_options2.extractRequestId;
      throwIf(!attachRequestId || !extractRequestId, "Please define 'options.attachRequestId / options.extractRequestId' for sending requests.");
    }
  }]);

  return ZitiWebSocket;
}();

module.exports = ZitiWebSocket;